/**
  *  \file client/scr-result.cc
  *  \brief Battle Result Screen
  *
  *  This contaisn the routines for starting a simulation and
  *  displaying its result summary.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/scr-result.h"

#include "client/vcrplay.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/event.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "ui/widget-busy.h"
#include "client/scr-sim.h"
#include "ui/table.h"
#include "defs.h"

#ifdef CONF_FLAK_SUPPORT
# include "flak/flak-db.h"
# include "flak/vis-ui.h"
#endif


/******************************** Playback *******************************/

// FIXME! This is almost cut&paste from playvcr.cc. It should probably be moved into a library.

/** Play VCRs. */
static void
playVcrs(GClassicVcrDatabase& db)
{
    ObjectHolder holder;
    WVcrSelector sel(db);
    WVcrSelectorWindow w(sel);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(holder.add(new UIButton(cm_VCR_Prev, "\036", '-')).addCommand());
    but.add(holder.add(new UIButton(cm_VCR_Next, "\037", '+')).addCommand());
    but.add(holder.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)).addCommand());
    but.add(holder.add(new UISpacer()));
    but.add(holder.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE)).addStop());

    w.setLayout(UIVBoxLayout::instance5);
    w.add(sel);
    w.add(but);
    w.pack();
    ui_root->centerWidget(w);
    ui_root->add(w);
    ui_root->run();
    ui_root->close(w);
}

class SimVcrVisitor : public GVcrDatabaseVisitor {
 public:
    void visitClassicDatabase(GClassicVcrDatabase& db)
        {
            playVcrs(db);
        }
    void visitFlakDatabase(GFlakVcrDatabase& db)
        {
#ifdef CONF_FLAK_SUPPORT
            if (db.getCount())
                playFlakBattle(db.getBattle(0));
#endif
        }
};

/************************** WSimClassResultList **************************/

WSimClassResultList::WSimClassResultList(GPlayerSet players, GSimResultSummary& result)
    : UIListbox(0, 15, 40 * font_ems[FONT_NORMAL], 0),
      result(result),
      players(players),
      desc_width(0),
      column_width(0)
{
    setHeaderHeight(getItemHeight());
}

WSimClassResultList::~WSimClassResultList()
{ }

void
WSimClassResultList::drawPart(GfxCanvas& can, int from, int to)
{
    UISkin& sk = getSkin();
    sk.startDraw();

    GfxParameters& parm = sk;
    for (int i = from; i < to; ++i) {
        const int me = getTop() + i;
        if (me >= result.getNumClassResults()) {
            drawEntry(can, UIColor::tc_Static, i, string_t());
        } else {
            const UIColorSet& scheme = (me == getCurrent() ? BLACK_COLOR_SET : GRAY_COLOR_SET);
            const int y = i * getItemHeight() + getHeaderHeight() + extent.y;
            const GSimResultSummary::ClassResult& class_result = result.getClassResult(me);

            parm.bgcolor = standard_colors[scheme[UIColor::tc_Background]];
            parm.color = standard_colors[scheme[UIColor::tc_Static]];
            parm.font = fonts[FONT_NORMAL];
            parm.just_x = 0;
            parm.just_y = 0;
            if (result.getTotalWeight() == 1) {
                outTextF(can, parm, extent.x, y, desc_width,
                         format("%d (%.1f%%)",
                                class_result.cumulative_weight,
                                100.0*class_result.cumulative_weight / result.getCumulativeWeight()));
            } else {
                outTextF(can, parm, extent.x, y, desc_width,
                         format("%.1f%%",
                                100.0*class_result.cumulative_weight / result.getCumulativeWeight()));
            }

            int x = extent.x + desc_width + column_width;
            parm.just_x = 2;
            for (int i = 1; i <= NUM_OWNERS; ++i) {
                if (players.contains(i)) {
                    if (class_result.owned_units[i] == 0)
                        outTextF(can, parm, x, y, column_width, "-");
                    else
                        outTextF(can, parm, x, y, column_width, numToString(class_result.owned_units[i]));
                    x += column_width;
                }
            }
            int xr = extent.x + extent.w;
            int xl = x - column_width;
            if (xl < xr)
                drawSolidBar(can, GfxRect(xl, y, xr - xl, font_heights[FONT_NORMAL]), parm.bgcolor);
        }
    }
}

void
WSimClassResultList::drawContent(GfxCanvas& can)
{
    UISkin& sk = getSkin();
    sk.startDraw();
    sk.getParameters().just_x = 0;
    sk.getParameters().just_y = 0;
    sk.setColor(UIColor::tc_Static);

    int x = extent.x + desc_width;
    for (int i = 1; i <= NUM_OWNERS; ++i) {
        if (players.contains(i)) {
            sk.outTextF(can, FONT_NORMAL, x, extent.y, column_width, getAdjName(i));
            x += column_width;
        }
    }

    UIListbox::drawContent(can);
}

void
WSimClassResultList::onResize(GfxRect old_size, bool by_layout)
{
    UIListbox::onResize(old_size, by_layout);

    // count number of players
    int num_players = 0;
    for (int i = 1; i <= NUM_OWNERS; ++i)
        if (players.contains(i))
            ++num_players;

    // compute sizes
    // FIXME: PCC 1.x does this differently
    if (num_players != 0) {
        int raw_desc_width = fonts[FONT_NORMAL]->getTextWidth("999x (100.0%)") + 10;
        if (raw_desc_width > extent.w)
            raw_desc_width = extent.w;

        int raw_col_width = (extent.w - raw_desc_width) / num_players;
        if (raw_col_width < 1)
            raw_col_width = 1;
        if (raw_col_width > font_ems[FONT_NORMAL] * 7)
            raw_col_width = font_ems[FONT_NORMAL] * 7;

        column_width = raw_col_width;
        desc_width   = raw_desc_width; // extent.w - raw_col_width * num_players;
    }
}

void
WSimClassResultList::updateFromResult()
{
    setSize(result.getNumClassResults());
    adjustPosition();
    drawWidget();
    sig_moved.raise(getId());
}

/************************* WSimUnitResultSummary *************************/

class WSimUnitResultSummary : public UIWidget {
    GSimResultSummary& summary;
    WSimList& list;
    SignalConnection conn_scroll;
 public:
    WSimUnitResultSummary(GSimResultSummary& summary, WSimList& list);
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    void onScroll();
};

WSimUnitResultSummary::WSimUnitResultSummary(GSimResultSummary& summary, WSimList& list)
    : UIWidget(),
      summary(summary),
      list(list),
      conn_scroll(list.sig_moved.add(this, &WSimUnitResultSummary::onScroll))
{ }

void
WSimUnitResultSummary::drawContent(GfxCanvas& can)
{
    const GSimResultSummary::UnitResult& res = summary.getUnitResult(list.getCurrent());

    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);
    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Static, _("Fought:"));
    tw.writeVariableCell(UIColor::tc_Static, _("Survived:"));
    tw.writeVariableCell(UIColor::tc_Static, _("Captured:"));

    tw.newRColumn(font_ems[FONT_NORMAL]*6);
    tw.writeVariableCell(UIColor::tc_Static, numToString(res.fought));
    tw.writeVariableCell(UIColor::tc_Static, numToString(res.won));
    tw.writeVariableCell(UIColor::tc_Static, numToString(res.captured));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Static, format(" (%.1f%%)", 100.0 * res.fought / summary.getCumulativeWeight()));
    tw.writeVariableCell(UIColor::tc_Static, format(" (%.1f%%)", 100.0 * res.won / summary.getCumulativeWeight()));
    tw.writeVariableCell(UIColor::tc_Static, format(" (%.1f%%)", 100.0 * res.captured / summary.getCumulativeWeight()));

    tw.finish();
    can.update(getExtent());
}

void
WSimUnitResultSummary::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = GfxPoint(15 * font_ems[FONT_NORMAL], 3 * font_heights[FONT_NORMAL]);
    info.flags = li_Flexible;
}

void
WSimUnitResultSummary::onScroll()
{
    drawWidget();
}

/************************* WSimUnitStat *************************/

class WSimUnitStat : public UIWidget {
    GSimResultSummary& summary;
    WSimList& list;
    const GSimState& initial_state;
    SignalConnection conn_scroll;
 public:
    WSimUnitStat(GSimResultSummary& summary, WSimList& list, const GSimState& initial_state);
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    void onScroll();
};

WSimUnitStat::WSimUnitStat(GSimResultSummary& summary, WSimList& list, const GSimState& initial_state)
    : UIWidget(),
      summary(summary),
      list(list),
      initial_state(initial_state),
      conn_scroll(list.sig_moved.add(this, &WSimUnitStat::onScroll))
{ }

void
WSimUnitStat::drawContent(GfxCanvas& can)
{
    const GSimResultSummary::UnitResult& res = summary.getUnitResult(list.getCurrent());

    static const char* const columns[] = {
        N_("Damage:"),
        N_("Shields:"),
        N_("Defense Lost:"),
        N_("SB Ftrs Lost:"),
        N_("Crew Left:"),
        N_("Fighters Lost:"),
        N_("Fighters Left:"),
        N_("Min Ftr aboard:"),
        N_("Torps Launched:"),
        N_("Torps Left:"),
        N_("Torps Hit:"),
    };
    enum { Damage, Shields,
           DefenseLost, SBFtrsLost,
           CrewLeft,
           FightersLost, FightersLeft, MinFtrAboard,
           TorpsLaunched, TorpsLeft, TorpsHit };

    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);

    std::vector<GSimStatItem> data;
    data.reserve(10);

    // display headings and populate data
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, "");
    for (unsigned int i = 0; i < countof(columns); ++i)
        tw.sizeCell(_(columns[i]));

    tw.writeVariableCell(UIColor::tc_Static, _(columns[Damage]));
    tw.writeVariableCell(UIColor::tc_Static, _(columns[Shields]));
    data.push_back(res.damage_taken);
    data.push_back(res.shields_left);

    if (summary.isPlanet(list.getCurrent())) {
        tw.writeVariableCell(UIColor::tc_Static, _(columns[DefenseLost]));
        tw.writeVariableCell(UIColor::tc_Static, _(columns[SBFtrsLost]));
        data.push_back(res.crew_left_or_defense_lost);
        data.push_back(res.fighters_lost);
    } else {
        const GSimShip& ship = initial_state.getShip(list.getCurrent());

        tw.writeVariableCell(UIColor::tc_Static, _(columns[CrewLeft]));
        data.push_back(res.crew_left_or_defense_lost);
        if (ship.getBayCount() > 0) {
            tw.writeVariableCell(UIColor::tc_Static, _(columns[FightersLost]));
            tw.writeVariableCell(UIColor::tc_Static, _(columns[FightersLeft]));
            tw.writeVariableCell(UIColor::tc_Static, _(columns[MinFtrAboard]));
            data.push_back(res.fighters_lost);
            data.push_back(GSimStatItem(res.fighters_lost, ship.getAmmo(), summary.getCumulativeWeight()));
            data.push_back(res.min_fighters_aboard);
        } else if (ship.getTorpLauncherCount() > 0) {
            tw.writeVariableCell(UIColor::tc_Static, _(columns[TorpsLaunched]));
            tw.writeVariableCell(UIColor::tc_Static, _(columns[TorpsLeft]));
            tw.writeVariableCell(UIColor::tc_Static, _(columns[TorpsHit]));
            data.push_back(res.torps_fired);
            data.push_back(GSimStatItem(res.torps_fired, ship.getAmmo(), summary.getCumulativeWeight()));
            data.push_back(res.torps_hit);
        }
    }

    // now show the actual numbers
    tw.newRColumn(3*font_ems[FONT_NORMAL]);
    tw.writeVariableCell(UIColor::tc_Static, _("min."));
    for (unsigned int i = 0; i < data.size(); ++i)
        tw.writeVariableCell(UIColor::tc_Static, numToString(data[i].getMin()));

    tw.newRColumn(4*font_ems[FONT_NORMAL]);
    tw.writeVariableCell(UIColor::tc_Static, _("average"));
    for (unsigned int i = 0; i < data.size(); ++i)
        tw.writeVariableCell(UIColor::tc_Static,
                             format("%.1f", double(data[i].getTotalScaled()) / summary.getCumulativeWeight()));

    tw.newRColumn(3*font_ems[FONT_NORMAL]);
    tw.writeVariableCell(UIColor::tc_Static, _("max."));
    for (unsigned int i = 0; i < data.size(); ++i)
        tw.writeVariableCell(UIColor::tc_Static, numToString(data[i].getMax()));

    tw.finish();
    can.update(getExtent());
}

void
WSimUnitStat::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(20, 6);
    info.flags = li_Flexible;
}

void
WSimUnitStat::onScroll()
{
    drawWidget();
}

/**************************** WSimResultWindow ***************************/

static GPlayerSet
countPlayers(const GSimState& st)
{
    GPlayerSet result;
    if (st.hasPlanet())
        result += st.getPlanet().getOwner();
    for (GSimState::ship_index_t i = 0; i < st.getNumShips(); ++i)
        result += st.getShip(i).getOwner();
    return result;
}

/** Add result buttons to widget structure.
    \param out Group to add to
    \param holder Space for new object
    \param default_button The card-specific button to use */
static void
addResultButtons(UIGroup& out, ObjectHolder& holder, UIButton& default_button)
{
    UIGroup& group_buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group_buttons2 = holder.add(new UIGroup(UIHBoxLayout::instance5));

    group_buttons.add(default_button);
    group_buttons.add(holder.add(new UISpacer()));
    group_buttons.add(holder.add(new UIButton(cm_Escape, _("ESC"), SDLK_ESCAPE)).addStop());

    group_buttons2.add(holder.add(new UIButton(_("Space - Again"), ' ')));
    group_buttons2.add(holder.add(new UIButton(_("S - Series"), 's')));
    group_buttons2.add(holder.add(new UIButton(_("R - Repeat"), 'r')));
    group_buttons2.add(holder.add(new UISpacer()));

    out.add(group_buttons2);
    out.add(group_buttons);
}

/** Constructor. */
WSimResultWindow::WSimResultWindow(const GSimState& sim_setup, const GSimOptions& sim_options)
    : UIWindow(_("Simulation Results"), BLUE_WINDOW, COLOR_GRAY),
      sim_setup(sim_setup),
      sim_options(sim_options),
      last_result(),
      result_summary(),
      class_result_list(countPlayers(sim_setup), result_summary),
      summary_line("", UIStaticText::BlackTitle),
      summary_line2("", UIStaticText::BlackTitle),
      tab_content(),
      tabs(tab_content),
      // FIXME: can we avoid the const_cast?
      result_list(holder.add(new WSimList(const_cast<GSimState&>(sim_setup), true)))
{
    /* Widget structure now looks like this:
         Window
           CardGroupTab
           CardGroup
             summary_tab
               list with scrollbar
               summary_line
               button bar
             result_tab
               list, info widget
               summary_line2
               button bar
       That is, we have two copies of the summary_line, and two copies
       of the buttons, although we only want to exchange one button, namely
       'Watch sample' vs 'Edit ship'. Possible alternatives:
       - rearrange stuff. Tried it, but I'm so used to the CCBSim 1.x layout
         that I think the changed one makes it worse.
       - change the label of the button on the fly. Currently not supported
         by the widget system; may make problems with layouting. */
    setLayout(UIVBoxLayout::instance5);

    UIGroup& summary_tab = holder.add(new UIGroup(UIVBoxLayout::instance5, 1000));
    UIGroup& result_tab  = holder.add(new UIGroup(UIVBoxLayout::instance5, 1001));

    UIGroup& summary_content = holder.add(new UIGroup(UIHBoxLayout::instance0));

    summary_content.add(class_result_list);
    summary_content.add(holder.add(new UIScrollbar(class_result_list)));
    summary_tab.add(summary_content);
    summary_tab.add(summary_line);

    UIGroup& result_content = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& result_rhs = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIFrameGroup& result_lhs = holder.add(new UIFrameGroup(UIHBoxLayout::instance0));

    WSimUnitResultSummary& result_sum1 = holder.add(new WSimUnitResultSummary(result_summary, result_list));
    WSimUnitStat& result_sum2 = holder.add(new WSimUnitStat(result_summary, result_list, sim_setup));
    result_lhs.add(result_list);
    result_lhs.add(holder.add(new UIScrollbar(result_list)));
    result_content.add(result_lhs);
    result_content.add(result_rhs);
    result_rhs.add(result_sum1);
    result_rhs.add(result_sum2);
    result_rhs.add(holder.add(new UISpacer()));

    result_tab.add(result_content);
    result_tab.add(summary_line2);

    addResultButtons(summary_tab, holder, holder.add(new UIButton(_("ENTER - Watch sample"), SDLK_RETURN)));
    addResultButtons(result_tab,  holder, holder.add(new UIButton(_("E - Edit this ship"), 'e')));

    tab_content.add(summary_tab);
    tab_content.add(result_tab);

    tabs.addTab(1000, UIKeyString(_("Totals")));
    tabs.addTab(1001, UIKeyString(_("Details")));

    add(tabs);
    add(tab_content);
    add(holder.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Horiz)));
    class_result_list.setFocusedWidget();

    pack();

    sig_run.add(&class_result_list, &WSimClassResultList::updateFromResult);
    sig_run.add(&result_sum1,       &WSimUnitResultSummary::onScroll);
    sig_run.add(&result_sum2,       &WSimUnitStat::onScroll);
}

WSimResultWindow::~WSimResultWindow()
{ }

/** Run simulation.
    \param series true if we're simulating a series; stop after one run
    \param max    maximum number of battles to generate, 0 for infinite

    \todo On SMP/Multicore machines, we could fork off some worker
    threads to do the runSimulation(). Currently, this does not work
    because our VCR algorithms may use static storage. */
void
WSimResultWindow::runSimulation(bool series, int max)
{
    ProgressMonitor monitor;
    UIBusyWidget busy_widget;
    ui_root->add(busy_widget);
    ui_root->performDeferredRedraws();
    monitor.sig_tick.add(&busy_widget, &UIBusyWidget::tick);

    long start = getTimer();
    while (1) {
        GSimState new_state(sim_setup);
        last_result.init(result_summary.getNumBattles());
        ::runSimulation(new_state, sim_options, last_result, monitor);
        result_summary.addResult(sim_setup, new_state, last_result);

        /* check exit conditions */
        if (max != 0) {
            if (--max == 0)
                break;
        }
        if (series) {
            if (last_result.series_length != 0)
                if ((result_summary.getNumBattles() % last_result.series_length) == 0)
                    break;
        }

        /* poll events */
        UIEvent event;
        do
            getEvent(event, true);
        while (event.type != ev_Key && event.type != ev_Idle);
        if (event.type == ev_Key && getClassOfKey(event.key.code) == KeyNormal)
            break;

        long now = getTimer();
        if (now > start + TIMER_HZ/2) {
            updateResults();
            ui_root->performDeferredRedraws();
            start = now;
        }
    }
}

/** Run first simulation.
    \return true on success, false if setup doesn't give any fights */
bool
WSimResultWindow::runFirstSimulation()
{
    ProgressMonitor monitor;

    GSimState new_state(sim_setup);
    last_result.init(result_summary.getNumBattles());
    ::runSimulation(new_state, sim_options, last_result, monitor);
    if (last_result.battles != 0 && last_result.battles->getCount() != 0) {
        result_summary.addResult(sim_setup, new_state, last_result);
        updateResults();
        return true;
    } else {
        return false;
    }
}

/** Update our widgets from the battle result. */
void
WSimResultWindow::updateResults()
{
    sig_run.raise();

    string_t result = format(_("%d simulation%!1{s%}"), result_summary.getNumBattles());
    summary_line.setText(result);
    summary_line2.setText(result);
}

bool
WSimResultWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_RETURN:
            if (tabs.getCurrentTab() == 1000) {
                SimVcrVisitor().visit(*result_summary.getClassResult(class_result_list.getCurrent()).sample_battle);
            }
            return true;
         case 'e':
            if (tabs.getCurrentTab() == 1001) {
                stop(cm_Enter);
            }
            return true;
         case ' ':
            runSimulation(false, 1);
            updateResults();
            return true;
         case 's':
            runSimulation(true, 0);
            updateResults();
            return true;
         case 'r':
            runSimulation(false, 0);
            updateResults();
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

int
WSimResultWindow::getResultListIndex() const
{
    return result_list.getCurrent();
}
