/**
  *  \file client/wscanchart.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wscanchart.h"
#include "gfx/clip.h"
#include "gfx/palette.h"
#include "game/lock.h"
#include "game/config.h"
#include "game/host.h"
#include "ui/uiglobal.h"

/************************** WScannerChartWidget **************************/

/*! \class WScannerChartWidget
    \brief Chart Widget for Control Screen Scanners

    This contains a standard chart widget with a movable scanner beam. */

/** Constructor. */
WScannerChartWidget::WScannerChartWidget()
    : scan(getViewport().center),
      conn_change(getDisplayedTurn().getCurrentUniverse().sig_universe_changed.add(static_cast<UIBaseWidget*>(this), (void (UIBaseWidget::*)()) &UIBaseWidget::drawWidget))
{ }

WScannerChartWidget::~WScannerChartWidget()
{ }

void
WScannerChartWidget::drawScannerXor(GfxCanvas& can)
{
    GfxXorFilter      xor_flt(can);
    GfxClipRectFilter can1(getExtent(), xor_flt);

    drawScanner(can1);
}

/** Draw end of scanner beam, large version.
    \param can Canvas
    \param parm Parameters
    \param pt Screen coordinates */
void
WScannerChartWidget::drawEndBig(GfxCanvas& can, GfxParameters& parm, GfxPoint pt)
{
    drawVLine(can, parm, pt.x-3, pt.y-2, pt.y+2);
    drawVLine(can, parm, pt.x+3, pt.y-2, pt.y+2);
    drawHLine(can, parm, pt.x-2, pt.y-3, pt.x+2);
    drawHLine(can, parm, pt.x-2, pt.y+3, pt.x+2);
}

/** Draw end of scanner beam, small version.
    \param can Canvas
    \param parm Parameters
    \param pt Screen coordinates  */
void
WScannerChartWidget::drawEndSmall(GfxCanvas& can, GfxParameters& parm, GfxPoint pt)
{
    can.drawPixel(pt.x+1, pt.y, parm.color, GFX_ALPHA_OPAQUE);
    can.drawPixel(pt.x-1, pt.y, parm.color, GFX_ALPHA_OPAQUE);
    can.drawPixel(pt.x, pt.y+1, parm.color, GFX_ALPHA_OPAQUE);
    can.drawPixel(pt.x, pt.y-1, parm.color, GFX_ALPHA_OPAQUE);
}

/** Draw scanner beam. */
void
WScannerChartWidget::drawScanner(GfxCanvas& can)
{
    GfxParameters parm;
    parm.color = standard_colors[COLOR_YELLOW];
    const GChartViewport& vp = getViewport();
    void (*f)(GfxCanvas&,GfxParameters&,GfxPoint) = vp.mult >= vp.divi ? drawEndBig : drawEndSmall;

    GfxPoint cen = vp.getCenter();
    GfxPoint end = vp.scaleXY(scan);

    f(can, parm, cen);
    if (end != cen) {
        drawLine(can, parm, cen.x, cen.y, end.x, end.y);
        f(can, parm, end);
    }
}

/** Query current location information for planet lock.
    \param hyperjumping        [out] true if we're a ship that is hyperjumping
    \param current_gravity_pid [out] Id of planet whose warp well we're in */
void
WScannerChartWidget::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    hyperjumping        = false;
    current_gravity_pid = (isValid()
                           ? getDisplayedTurn().getCurrentUniverse().getGravityPlanetAt(getCenterPosition())
                           : 0);
}

/** Query distance to a point. Returns any distance metric. Can return
    -1 to reject a point. */
int32
WScannerChartWidget::lockQueryDistance(GPoint pt)
{
    if (isValid())
        return distanceSquared(getCenterPosition(), pt);
    else
        return -1;
}

/** Draw below starchart. This draws the center cross. */
void
WScannerChartWidget::drawPre(GfxCanvas& can)
{
    const GfxPoint m = getViewport().getCenter();

    GfxParameters parm;
    parm.color = standard_colors[COLOR_BLUEBLACK];

    drawHLine(can, parm, m.x - 10, m.y,      m.x + 11);
    drawVLine(can, parm, m.x,      m.y - 10, m.y + 11);
}

/** Draw above starchart. This draws the scanner. */
void
WScannerChartWidget::drawPost(GfxCanvas& can)
{
    if (isValid())
        drawScannerXor(can);
}

/** Clip scanner position.
    \return pt modified such that it is in the visible range
    \todo maybe move to WChartWidget? */
GPoint
WScannerChartWidget::clipScannerPosition(GPoint pt) const
{
    const GChartViewport& vp = getViewport();
    const GPoint m           = vp.center;
    const int xw = vp.unscale(extent.w) / 2;
    const int yw = vp.unscale(extent.h) / 2;
    if (pt.x < m.x - xw)
        pt.x = m.x - xw;
    if (pt.x > m.x + xw)
        pt.x = m.x + xw;
    if (pt.y < m.y - yw)
        pt.y = m.y - yw;
    if (pt.y > m.y + yw)
        pt.y = m.y + yw;
    return pt;
}

/** Set scanner position.
    \param pt      point
    \param limit   true to limit actual value to visible range,
                   false to take new value without checking */
void
WScannerChartWidget::setScannerPosition(GPoint pt)
{
    if (pt != scan) {
        if (hasState(st_Exposed) && !hasState(st_WantRedraw)) {
            /* widget is visible and up-to-date */
            ObjectHolder h;
            GfxCanvas& can = getParent()->getCanvasFor(*this, h);
            can.lock();
            drawScannerXor(can);
            scan = pt;
            drawScannerXor(can);
            can.unlock();
            can.update(extent);
            sig_moved.raise(getId());
        } else {
            /* widget is either not visible, or will be redrawn soon,
               so we don't need to do anything */
            scan = pt;
            sig_moved.raise(getId());
        }
    }
}

/** Change scanner center position. This will also reset the scanner
    position to point to the center. */
void
WScannerChartWidget::setCenterPosition(GPoint pt)
{
    WChartWidget::setCenterPosition(pt);
    setScannerPosition(pt);
}

void
WScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    setScannerPosition(pt);
}

/** Move according to a keypress. */
void
WScannerChartWidget::moveKey(int dx, int dy)
{
    setSelectedWidget();
    GPoint pt = getScannerPosition();
    userMoveScannerTo(clipScannerPosition(GPoint(pt.x + dx, pt.y + dy)));
}

/** Handle keypress.
    \param k key
    \return true iff key handled */
bool
WScannerChartWidget::handleKey(UIEvent::Key k)
{
    switch (k.code) {
     case SDLK_LEFT:
        moveKey(k.arg ? -k.arg : -10, 0);
        return true;
     case SDLK_RIGHT:
        moveKey(k.arg ? k.arg : 10, 0);
        return true;
     case SDLK_UP:
        moveKey(0, k.arg ? k.arg : 10);
        return true;
     case SDLK_DOWN:
        moveKey(0, k.arg ? -k.arg : -10);
        return true;
     case ss_Shift + SDLK_LEFT:
        moveKey(-1, 0);
        return true;
     case ss_Shift + SDLK_RIGHT:
        moveKey(1, 0);
        return true;
     case ss_Shift + SDLK_UP:
        moveKey(0, 1);
        return true;
     case ss_Shift + SDLK_DOWN:
        moveKey(0, -1);
        return true;
     case ss_Ctrl + SDLK_LEFT:
        moveKey(-100, 0);
        return true;
     case ss_Ctrl + SDLK_RIGHT:
        moveKey(100, 0);
        return true;
     case ss_Ctrl + SDLK_UP:
        moveKey(0, 100);
        return true;
     case ss_Ctrl + SDLK_DOWN:
        moveKey(0, -100);
        return true;
     case '+':
        if (hasState(sc_Zoomable))
            zoomIn();
        return true;
     case '-':
        if (hasState(sc_Zoomable))
            zoomOut();
        return true;
     case ' ':                               doItemLock(getScannerPosition(), lm_Left,  false, false); return true;
     case SDLK_RETURN:                       doItemLock(getScannerPosition(), lm_Right, false, false); return true;
     case ss_Ctrl + ' ':                     doItemLock(getScannerPosition(), lm_Left,  true,  false); return true;
     case ss_Ctrl + SDLK_RETURN:             doItemLock(getScannerPosition(), lm_Right, true,  false); return true;
     case ss_Shift + ' ':                    doItemLock(getScannerPosition(), lm_Left,  false, true);  return true;
     case ss_Shift + SDLK_RETURN:            doItemLock(getScannerPosition(), lm_Right, false, true);  return true;
     case ss_Shift + ss_Ctrl + ' ':          doItemLock(getScannerPosition(), lm_Left,  true,  true);  return true;
     case ss_Shift + ss_Ctrl + SDLK_RETURN:  doItemLock(getScannerPosition(), lm_Right, true,  true);  return true;

        /* z
           m c-m */
    }
    return false;
}

/** Lock at item. This adjusts the clicked point, and calls userMoveScannerTo()
    with the new position.
    \param clicked       Point clicked by user
    \param mode          Lock mode (left/right mouse button)
    \param marked_only   true to accept only marked units
    \param optimize_warp true to optimize using gravity wells */
void
WScannerChartWidget::doItemLock(GPoint clicked,
                                TLockMode mode,
                                bool marked_only,
                                bool optimize_warp)
{
    // Regular lock code
    const GChartViewport& vp = getViewport();
    const GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GLockData locker(clicked,
                     GPoint(vp.unscaleX(extent.x),
                            vp.unscaleY(extent.y + extent.h - 1)),
                     GPoint(vp.unscaleX(extent.x + extent.w - 1),
                            vp.unscaleY(extent.y)),
                     marked_only);

    if ((getLockMode(mode) & li_Planet) != 0 && optimize_warp) {
        // Optimize warp wells
        findPlanet(univ, locker);
        GPoint found_point = locker.getPoint();

        // Query current position
        bool hyperjumping;
        int current_pid;
        lockQueryLocation(hyperjumping, current_pid);

        // Can we optimize warp wells?
        int clicked_pid = univ.getPlanetAt(found_point);
        if (clicked_pid > 0
            && config.AllowGravityWells()
            && (!hyperjumping || !host.isPHost() || config.AllowHyperjumpGravWells())
            && clicked_pid != current_pid)
        {
            /* We try to find the edge of a gravity well unless
               - we're heading for deep space, i.e. no planet found
               - gravity wells are disabled
               - we're starting inside the same gravity well we started in,
                 in this case we assume we want to move to the planet */
            const int wwrange = (host.isPHost() ? config.GravityWellRange() :
                                 hyperjumping ? 2 : 3);

            // Start with the assumption that moving directly is the best choice.
            // Then try all points in warp well range.
            int32 best_dist   = lockQueryDistance(found_point);
            GPoint best_point = found_point;
            for (int dx = -wwrange; dx <= wwrange; ++dx) {
                for (int dy = -wwrange; dy <= wwrange; ++dy) {
                    GPoint new_point = GPoint(found_point.x + dx, found_point.y + dy);
                    int32 new_dist = lockQueryDistance(new_point);
                    if (new_dist >= 0
                        && (best_dist < 0 || new_dist < best_dist)
                        && univ.getGravityPlanetAt(new_point) == clicked_pid)
                    {
                        // Accept new point if it is valid, has a better metric than
                        // the previous one, and it is in the same warp well.
                        best_dist = new_dist;
                        best_point = new_point;
                    }
                }
            }

            // Move to found point, if any
            if (best_dist >= 0)
                userMoveScannerTo(best_point);
        } else {
            // No warp wells, so just move to found point
            userMoveScannerTo(found_point);
        }
    } else {
        // Regular locking only
        findItem(univ, locker, mode);
        userMoveScannerTo(locker.getPoint());
    }
}

bool
WScannerChartWidget::handleEvent(const UIEvent& e, bool second_pass)
{
    /* nothing selected -> no input accepted */
    if (!isValid())
        return WChartWidget::handleEvent(e, second_pass);

    if (e.type == ev_Mouse && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        const GChartViewport& vp = getViewport();
        setSelectedWidget();
        if (e.mouse.b || hasState(sc_KeyboardMode)) {
            if (e.mouse.b & mb_Double) {
                sig_dblclick.raise(getId());
            } else {
                if (e.mouse.b & mb_Left)
                    doItemLock(vp.unscaleXY(GfxPoint(e.mouse.x, e.mouse.y)),
                               lm_Left,
                               (e.mouse.b & ss_Ctrl) != 0,
                               (e.mouse.b & ss_Shift) != 0);
                else if (e.mouse.b & mb_Right)
                    doItemLock(vp.unscaleXY(GfxPoint(e.mouse.x, e.mouse.y)),
                               lm_Right,
                               (e.mouse.b & ss_Ctrl) != 0,
                               (e.mouse.b & ss_Shift) != 0);
                else
                    userMoveScannerTo(vp.unscaleXY(GfxPoint(e.mouse.x, e.mouse.y)));
            }
        }
        return true;
    } else if (e.type == ev_Key && hasState(sc_KeyboardMode | sc_Keyboard)) {
        return handleKey(e.key);
    } else {
        return WChartWidget::handleEvent(e, second_pass);
    }
}

/***************************** Keyboard Mode *****************************/

// FIXME: this is a copy of the same routine in widget-busy.cc
static void
drawActivityFrame(GfxCanvas& can, const GfxRect& r, uint8 p)
{
    const int x0 = r.x;
    const int x1 = r.x + r.w - 1;
    const int y0 = r.y;
    const int y1 = r.y + r.h - 1;

    const uint32 color1 = standard_colors[COLOR_YELLOW];
    const uint32 color0 = standard_colors[COLOR_DARK];

    int x = x0, y = y0;
    /* top-left to top-right */
    while (x < x1) {
        can.drawPixel(x, y, ((p & 7) < 4) ? color0 : color1, GFX_ALPHA_OPAQUE);
        ++p;
        ++x;
    }
    /* top-right to bottom-right */
    while (y < y1) {
        can.drawPixel(x, y, ((p & 7) < 4) ? color0 : color1, GFX_ALPHA_OPAQUE);
        ++p;
        ++y;
    }
    /* bottom-right to bottom-left */
    while (x > x0) {
        can.drawPixel(x, y, ((p & 7) < 4) ? color0 : color1, GFX_ALPHA_OPAQUE);
        ++p;
        --x;
    }
    /* bottom-left to top-left */
    while (y > y0) {
        can.drawPixel(x, y, ((p & 7) < 4) ? color0 : color1, GFX_ALPHA_OPAQUE);
        ++p;
        --y;
    }
}

namespace {
    class KeyboardModeHelper {
        WScannerChartWidget& w;
     public:
        KeyboardModeHelper(WScannerChartWidget& w);
        ~KeyboardModeHelper();

        void drawFrame(uint8 phase);
        void clearFrame();
    };
}

KeyboardModeHelper::KeyboardModeHelper(WScannerChartWidget& w)
    : w(w)
{
    w.setState(w.sc_Keyboard, true);
    w.setSelectedWidget();
}

KeyboardModeHelper::~KeyboardModeHelper()
{
    w.setState(w.sc_Keyboard, false);
    clearFrame();
}

void
KeyboardModeHelper::drawFrame(uint8 phase)
{
    // Obtain canvas and size
    ObjectHolder tmp;
    GfxCanvas& can = w.getParent()->getCanvasFor(w, tmp);
    GfxRect frame_size = w.getExtent();
    frame_size.grow(2, 2);

    // Draw
    can.lock();
    drawActivityFrame(can, frame_size, phase);
    can.unlock();
    can.update(frame_size);
}

void
KeyboardModeHelper::clearFrame()
{
    GfxRect frame_size = w.getExtent();
    frame_size.grow(2, 2);
    ui_root->markRedrawRegion(frame_size);
}

/** Perform keyboard movement mode. This gives the WScannerChartWidget
    exclusive focus, and lets the user move the pointer using the
    keyboard or the mouse. This function returns when the user exits
    keyboard mode. */
void
WScannerChartWidget::doKeyboardMode()
{
    // Keyboard mode not available when there's no valid position
    if (!isValid())
        return;

    // Avoid recursion
    if (hasState(sc_Keyboard))
        return;

    // Regular keyboard mode
    KeyboardModeHelper helper(*this);
    uint8 frame_phase = 0;
    bool need_update_frame = true;
    bool running = true;
    int pending_dx = 0;
    int pending_dy = 0;
    while (running) {
        // Perform updates
        ui_root->performDeferredRedraws();
        if (need_update_frame) {
            helper.drawFrame(frame_phase/4);
            need_update_frame = false;
        }

        // Handle events
        UIEvent e;
        getEvent(e, false, true);
        switch (e.type) {
         case ev_Tick:
            // Slow down a little
            if ((++frame_phase & 3) == 0)
                need_update_frame = true;
            break;
         case ev_Key:
            /* PCC 1's exit commands:
                 1..9         prefix arg
                 BS, ESC, y   regular exit
                 x            exit + 'x'
                 C-w          exit + C-w
                 l            exit + 'l'
               We add some more. */
            switch (e.key.code) {
             case SDLK_ESCAPE:
             case SDLK_BACKSPACE:
             case 'y':
                running = false;
                break;

             case SDLK_F1 + ss_Ctrl:
             case SDLK_F2 + ss_Ctrl:
             case SDLK_F3 + ss_Ctrl:
             case SDLK_F4 + ss_Ctrl:
             case SDLK_F5 + ss_Ctrl:
             case 'x':
             case 'w' + ss_Ctrl:
             case 'l':
                running = false;
                ungetEvent(e);
                break;
             default:
                handleKey(e.key);
                break;
            }
            break;
         case ev_Mouse:
            {
                // Perform relative mouse movement
                pending_dx += e.mouse.x;
                pending_dy += e.mouse.y;
                GPoint new_pt = GPoint(scan.x + getViewport().unscale(pending_dx),
                                       scan.y - getViewport().unscale(pending_dy));
                if (new_pt != scan)
                    pending_dx = pending_dy = 0;

                // Find new position, by locking if needed
                new_pt = clipScannerPosition(new_pt);
                if (e.mouse.b & mb_Left) {
                    doItemLock(new_pt, lm_Left, (e.mouse.b & ss_Ctrl) != 0, (e.mouse.b & ss_Shift) != 0);
                } else if (e.mouse.b & mb_Right) {
                    doItemLock(new_pt, lm_Right, (e.mouse.b & ss_Ctrl) != 0, (e.mouse.b & ss_Shift) != 0);
                } else {
                    userMoveScannerTo(new_pt);
                }

                // Double-click exits
                if (e.mouse.b & mb_Double) {
                    running = false;
                }
            }
            break;
         case ev_Idle:
            break;
        }
    }
}

void
WScannerChartWidget::getLayoutInfo(LayoutInfo& info)
{
    // FIXME? Probably preferred size should be different
    info.flags = li_Flexible;
    info.min_size = info.pref_size = GfxPoint(100,100);
}
