/**
  *  \file flak-port.h
  *  \brief Infrastructure for Porting FLAK to Different Environments
  *
  *  While the FLAK core was written to fit into the PCC II kernel,
  *  care has been taken to isolate all environment access to ease
  *  replacing the actual back-end. If you want to run the FLAK core
  *  inside a different environment, you only have to provide a very
  *  thin "PCC II emulation layer".
  *
  *  This file provides an implementation of the porting layer to link
  *  FLAK against the PDK.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef FLAK_PORT_H_INCLUDED
#define FLAK_PORT_H_INCLUDED

#ifdef FLAK_USE_PDK
/*
 *    -- FLAK on top of the PDK --
 */
#include "phostpdk.h"

// Weapons. The core just needs access to weapon specs. Actually, the
// names GBeam and GTorpLauncher are not referenced, either; we use
// the accessor functions below.
class GBeam {
    int nr;
 public:
    GBeam(int nr) : nr(nr) { }
    inline int getKill() const { return BeamKillPower(nr); }
    inline int getDamage() const { return BeamDestructivePower(nr); }
};

class GTorpLauncher {
    int nr;
 public:
    GTorpLauncher(int nr) : nr(nr) { }
    inline int getKill() const { return TorpKillPower(nr); }
    inline int getDamage() const { return TorpDestructivePower(nr); }
};

inline GBeam         getBeam(int nr) { return GBeam(nr); }
inline GTorpLauncher getTorp(int nr) { return GTorpLauncher(nr); }
inline int           raceId(int player) { return gPconfigInfo->PlayerRace[player]; }

/* Get value of (scalar) config option 'xx' */
#define CONFIG_SCALAR(xx)   (::gPconfigInfo->xx)

/* Get player 'ii's value of array config option 'xx' */
#define CONFIG_ARRAY(xx,ii) (::gPconfigInfo->xx[ii])

/* Get experience-modified value of an option. You probably only have
   to replace the access to the EModXXX option. Parameters are
   KEY (the option), LEVEL (experience level, 0..10), PL (player),
   MIN/MAX (permitted range) */
#define CONFIG_EX(KEY, LEVEL, PL, MIN, MAX) \
   limit(CONFIG_ARRAY(KEY, PL) + ((LEVEL) ? ::gPconfigInfo->EMod##KEY[LEVEL-1] : 0), MIN, MAX)

enum {
    NUM_OWNERS  = 12,
    NUM_PLAYERS = 11,
    NUM_SHIPS   = SHIP_NR,
    NUM_PLANETS = PLANET_NR,
    NUM_TORPS   = TORP_NR,
    NUM_BEAMS   = BEAM_NR
};

#else /* FLAK_USE_PDK */
/*
 *    -- FLAK on top of PCC II Kernel --
 */

#include "game/spec.h"
#include "game/config.h"
#include "game/player.h"
#include "game/struct.h"

#define CONFIG_SCALAR(xx)      (::config.xx())
// CONFIG_ARRAY is defined in game/config.h

/** Get experience-modified value for an option. */
#define CONFIG_EX(KEY, LEVEL, PL, MIN, MAX) \
   limit(CONFIG_ARRAY(KEY, PL) + ((LEVEL) ? ::config.EMod##KEY(LEVEL) : 0), MIN, MAX)

#endif

#endif
