/**
  *  \file game/ufo.h
  *  \brief Ufo and Wormhole handling
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_UFO_H
#define PCC_V2_GAME_UFO_H

#include <list>
#include "game/obj.h"
#include "cpluslib/string.h"
#include "game/objtypes.h"
#include "game/utildata.h"

struct TUfo;
struct TDbUfo;
struct TUtil14Wormhole;
struct TUtil33GO;

class GUfo : public GCircularObject {
 public:
    GUfo(int id);
    GUfo(const GUfo& other);
    ~GUfo();

    GUfo& operator=(const GUfo& rhs);

    /* GCircularObject interface */
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;
    GPoint   getPos() const;
    int      getRadius() const;

    /* Loader interface */
    void addUfoData(const TUfo& data);
    void addObjectData(const TUtil33GO& obj);
    void addHistoryData(const TDbUfo& data);
    void addWormholeData(const TUtil14Wormhole& data, bool all, int turn);

    void connectWith(GUfo& other);
    void postprocess(int turn);

    bool isStoredInHistory() const;
    void getHistoryData(TDbUfo& data) const;

    bool isSeenThisTurn() const;

    /* Accessors */
    int      getColorCode() const;
    uint8    getColor() const;
    int      getSpeed() const;
    int      getHeading() const;
    int      getPlanetRange() const;
    int      getShipRange() const;
    int      getTypeCode() const;
    string_t getInfo1() const;
    string_t getInfo2() const;
    int32    getRealId() const;
    GPoint   getOtherEndPos() const;
    int      getLastTurn() const;
    GPoint   getMovementVector() const;

    /* Values for flags */
    enum {
        UfoSeenThisTurn     = 1,
        UfoStoredInDatabase = 2
    };

 private:
    /* Standard Ufo properties */
    int16  id;
    int16  vga_color;
    GPoint pos;
    int16  warp;
    int16  heading;
    int16  planet_range, ship_range;
    int16  radius;
    int16  type_code;
    string_t name;
    string_t info1, info2;

    /* Additional PCC properties */
    int32  real_id;
    int16  turn_last_seen;
    GPoint pos_last_seen;
    GPoint movement_vector;

    int    flags;

    GPoint target;
    // reference to target Ufo?

    void copyUfo(const TUfo& data);
};

class GUfoContainer {
    typedef std::list<GUfo> ufolist_t;
    ufolist_t ufos;
    int32 num_ufos;

    typedef std::list<TUtil14Wormhole> wormlist_t;
    wormlist_t wormholes;
 public:
    typedef ufolist_t::iterator iterator;
    typedef ufolist_t::const_iterator const_iterator;

    GUfoContainer();
    ~GUfoContainer();

    // Loading:
    void addWormholeData(const TUtil14Wormhole& wh);
    void addUfoData(int id, const TUfo& ufo);
    void addObjectData(const TUtil33GO& obj);
    void addHistoryData(const TDbUfo& ufo);
    void postprocess(GUniverse& univ);

    // Access:
    iterator begin();
    iterator end();
    const_iterator begin() const;
    const_iterator end() const;

    iterator getFirstAt(int id);

    GUfo& getUfoById(int id);

    // Indexed access:
    int32 getNumUfos() const;
    GUfo& getUfoByIndex(int32 index);
};

class GUfoType : public GObjectType {
 public:
    GUfoType();
    ~GUfoType();

    void init();
    bool isExistant(int index);
    GUfo& getObject(int index);
    int getNext(int index, bool marked);
    int getPrev(int index, bool marked);

    GUfo& getUfo()
        { return getObject(getCurrent()); }
};

inline bool
GUfo::isStoredInHistory() const
{
    return (flags & UfoStoredInDatabase) != 0;
}

inline bool
GUfo::isSeenThisTurn() const
{
    return (flags & UfoSeenThisTurn) != 0;
}

inline GUfoContainer::iterator
GUfoContainer::begin()
{
    return ufos.begin();
}

inline GUfoContainer::iterator
GUfoContainer::end()
{
    return ufos.end();
}

inline GUfoContainer::const_iterator
GUfoContainer::begin() const
{
    return ufos.begin();
}

inline GUfoContainer::const_iterator
GUfoContainer::end() const
{
    return ufos.end();
}

#endif
