/**
  *  \file game/vcrdb.h
  *  \brief VCR Database
  *
  *  As of early 2005, we support not only the HOST / PVCR
  *  distinction, but also the FLAK multi-ship combat engine which has
  *  a completely different VCR format. This module provides the
  *  database base classes, as well as the storage engine for
  *  "classic" (one-on-one) VCRs.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_VCRDB_H
#define PCC_V2_GAME_VCRDB_H

#include <vector>
#include "game/vcr.h"
#include "io/streams.h"

class GClassicVcrDatabase;
class GFlakVcrDatabase;
class GVcrDatabase;

/** VCR Database Visitor. Used as a type-switch to distinguish between
    different VCR database formats. */
class GVcrDatabaseVisitor : public Object {
 public:
    GVcrDatabaseVisitor()
        { }
    virtual ~GVcrDatabaseVisitor()
        { }

    virtual void visitClassicDatabase(GClassicVcrDatabase&) = 0;
    virtual void visitFlakDatabase(GFlakVcrDatabase&) = 0;

    void visit(GVcrDatabase& db);
};

/** Basic VCR database. */
class GVcrDatabase : public Object {
 public:
    GVcrDatabase()
        { }
    ~GVcrDatabase()
        { }

    /** Get number of battles in this database. */
    virtual int  getCount() const = 0;
    /** Visit this database. */
    virtual void visit(GVcrDatabaseVisitor&) = 0;
};

/****************************** Classic VCR ******************************/

class GVcrEntry {
    int  status;
    int  hulls[2];
    TVcr vcr;
    TVcr result;

    int guessHull(int side) const;
 public:
    GVcrEntry(const TVcr& vcr);
    ~GVcrEntry();

    void setResultFromPlayer(VcrPlayer& player);
    void setResult(int status, const TVcr& result);

    /** Get combat recording in "raw" format. */
    const TVcr& getVcr() const
        { return vcr; }
    /** Get result of combat in "raw" format. */
    const TVcr& getResult() const
        { return result; }
    /** Get status of fight, if known. */
    int getStatus() const
        { return status; }
    /** Get guessed hull, if known.
        \return hull number, zero if not known. */
    const int getGuessedHull(int side) const
        { return hulls[side]; }

    bool canBeHull(int side, int hull_id) const;

    /** Check whether \c side is a planet. */
    bool isPlanet(int side) const
        { return vcr.battle_type && side; }
};

class GClassicVcrDatabase : public GVcrDatabase {
 public:
    enum Type {
        vcr_Host,               ///< Host.
        vcr_Unknown,            ///< Unknown.
        vcr_UnknownPHost,       ///< Unknown brand of PHost.
        vcr_PHost2,             ///< PHost 2.
        vcr_PHost3,             ///< PHost 3 or PHost 4 without extensions.
        vcr_PHost4              ///< PHost 4 with extensions.
    };

    GClassicVcrDatabase(Type type);
    GClassicVcrDatabase(Stream& file);
    ~GClassicVcrDatabase();

    int  getCount() const;
    void visit(GVcrDatabaseVisitor&);

    GVcrEntry& getEntry(int n);
    GVcrEntry& addEntry(const TVcr& vcr);

    Ptr<VcrPlayer> createPlayer();

    uint16 getCapabilities()
        { return capabilities; }

    Type getType() const
        { return type; }
    static const char* getTypeName(Type t);

 private:
    typedef std::vector<GVcrEntry*> EntryVector;

    EntryVector content;
    Type        type;
    uint16      capabilities;
};


inline void
GVcrDatabaseVisitor::visit(GVcrDatabase& db)
{
    db.visit(*this);
}

#endif
