/**
  *  \file int/data.cc
  *  \brief Interpreter Data Structures
  *
  *  This defines the data model used in the interpreter.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "int/data.h"
#include "util/misc.h"
#include "cpluslib/assert.h"
#include "io/streams.h"
#include "io/memstream.h"
#include "io/io.h"
#include "util/except.h"

/** Create blank data segment. */
IntDataSegment::IntDataSegment()
{ }

/** Destroy. */
IntDataSegment::~IntDataSegment()
{
    safeClear(values);
}

/** Make index accessible. Makes sure that \c values can be indexed
    with \c index. */
void
IntDataSegment::makeIndexAccessible(uint32 index)
{
    if (index >= values.size())
        values.insert(values.end(), index - values.size() + 1, (IntValue*) 0);
}

/** Fetch element. Returns index \c index, or 0 if that index has never
    been written. */
IntValue*
IntDataSegment::operator[](uint32 index)
{
    if (index < values.size())
        return values[index];
    else
        return 0;
}

/** Set element by copying.
    \param index Index to store
    \param value Value to store, may be 0. Will be copied (IntValue::clone()) */
void
IntDataSegment::set(uint32 index, IntValue* value)
{
    if (value)
        value = value->clone();
    setNew(index, value);
}

/** Set element to newly-allocated value.
    \param index Index to store
    \param value Value to store, may be 0. The value will become owned by
                 the IntDataSegment object. If the assignment results in
                 an exception, it will be deleted. */
void
IntDataSegment::setNew(uint32 index, IntValue* value)
{
    try {
        makeIndexAccessible(index);
        delete values[index];
        values[index] = value;
    }
    catch (...) {
        delete value;
        throw;
    }
}

/** Append value by copying.
    \param value Value to append, may be 0. Will be copied (IntValue::clone()) */
void
IntDataSegment::push(IntValue* value)
{
    set(values.size(), value);
}

/** Append newly-allocated value.
    \param value Value to store, may be 0. The value will become owned by
                 the IntDataSegment object. If the assignment results in
                 an exception, it will be deleted. */
void
IntDataSegment::pushNew(IntValue* value)
{
    setNew(values.size(), value);
}

/** Remove last element. */
void
IntDataSegment::pop()
{
    ASSERT(values.size());
    delete values.back();
    values.pop_back();
}

/** Get top-of-stack element.
    \param index Element index. 0 (default) returns the last pushed element,
                 1 the one before, and so on.
    \pre index < size() */
IntValue*
IntDataSegment::top(uint32 index)
{
    ASSERT(index < values.size());
    return values[values.size() - index - 1];
}

/** Swap data segment elements.
    \param index Element index in this data segment.
    \param other Other data segment
    \param otherIndex Element index in other data segment. */
void
IntDataSegment::swap(uint32 index, IntDataSegment& other, uint32 otherIndex)
{
    makeIndexAccessible(index);
    makeIndexAccessible(otherIndex);
    std::swap(values[index], other.values[otherIndex]);
}

/** Clear this data segment. Discards all elements. */
void
IntDataSegment::clear()
{
    safeClear(values);
    std::vector<IntValue*>().swap(values);
}

/** Get size of data segment. Returns number of allocated elements. */
uint32
IntDataSegment::size() const
{
    return values.size();
}

/** Get number of used slots. Returns the smallest possible number such that all
    slots at higher numbers yield null with operator[]. */
uint32
IntDataSegment::getNumberOfUsedSlots()
{
    uint32 result = values.size();
    while (result > 0 && values[result-1] == 0)
        --result;
    return result;
}

/** Save to stream.
    \param out Stream to store in
    \param slots Number of slots to store
    \return true on success, false (or exception) if not enough room */
bool
IntDataSegment::save(Stream& out, uint32 slots)
{
    // Collect headers in one stream, aux data in another
    MemoryStream headers, aux;

    for (uint32 i = 0; i < slots; ++i) {
        // Generate single entry
        IntTagNode sv;
        if (IntValue* v = operator[](i)) {
            v->store(sv, aux);
        } else {
            sv.tag   = sv.Tag_Empty;
            sv.value = 0;
        }

        // Serialize tag node
        storeStructure(headers, sv);
    }

    // Generate output
    return out.write(headers.getBuffer().ptr(), headers.getBuffer().getSize()) == headers.getBuffer().getSize()
        && out.write(aux.getBuffer().ptr(),     aux.getBuffer().getSize())     == aux.getBuffer().getSize();
}

/** Load from stream.
    \param in Stream to load from
    \param slots Number of slots to load (slots with higher numbers will not be touched) */
void
IntDataSegment::load(Stream& in, uint32 slots)
{
    uint32 headerSize = slots * IntTagNode::size;
    Array<char> headers(headerSize);

    // Read headers
    in.readT(headers.ptr(), headerSize);

    // Read elements
    for (uint32 i = 0; i < slots; ++i) {
        // Read tag node
        IntTagNode sv;
        getStructure(headers.ptr() + (IntTagNode::size * i), sv);

        // Generate node
        set(i, IntValue::load(sv, in));
    }
}
