/**
  *  \file tools/c2sweep.cc
  *  \brief "Sweep" utility - Main Function
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include "game/sweep.h"
#include "util/console.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/misc.h"
#include "version.h"
#include "game/gamedirs.h"
#include "game/struct.h"

#define VERSION PCC2_VERSION

static const char* progname;

/** Exit with error. */
static void
die(string_t str)
{
    std::cerr << progname << ": " << str << std::endl;
    std::exit(1);
}

static void
help()
{
    std::cout << format(_("PCC2 Game Directory Cleaner v%s - (c) 2009 Stefan Reuther\n"), VERSION)
              << format(_("\n"
                          "Usage:\n"
                          "  %s [-h]\n"
                          "  %$0s [-nlx] [-a|player numbers] [gamedir [rootdir]]\n\n"
                          "Options:\n"
                          "  -n       Dry run, list files that would be deleted\n"
                          "  -l       Also erase database/log files usually kept several turns\n"
                          "  -a       Erase all players' files\n"
                          "  -x       Increase verbosity\n"
                          "\n"
                          "Report bugs to <Streu@gmx.de>\n"),
                        progname);
    std::exit(0);
}

namespace {
    class LoggingSweepProcessor : public GSweepProcessor {
     public:
        void removeFile(string_t name)
            {
                Ptr<Stream> s = game_file_dir->openFileNT(name, Stream::C_READ);
                if (s)
                    std::cout << name << "\n";
            }
    };
}

int
main(int /*argc*/, char** argv)
{
    GSweepRunProcessor processor;
    GSweepControl controller;

    bool opt_dry_run = false;

    const char* gameDir = 0;
    const char* rootDir = 0;

    progname = *argv;
    initNLS();
    enableStdioConsole();

    while (const char* arg = *++argv) {
        if (*arg == '-') {
            while (*++arg) {
                switch (*arg) {
                 case 'n':
                    opt_dry_run = true;
                    break;
                 case 'l':
                    controller.setEraseDatabase(true);
                    break;
                 case 'x':
                    processor.setVerbose(true);
                    break;
                 case 'a':
                    controller.setPlayers(GPlayerSet::allUpTo(NUM_PLAYERS));
                    break;
                 case 'h':
                    help();
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        } else {
            int n;
            if (stringValue(arg, n) && n > 0 && n <= NUM_PLAYERS) {
                /* Player number */
                controller.setPlayers(controller.getPlayers() + n);
            } else if (!gameDir) {
                /* Game directory */
                gameDir = arg;
            } else if (!rootDir) {
                /* Root directory */
                rootDir = arg;
            } else {
                die(_("too many arguments"));
            }
        }
    }

    if (controller.getPlayers().empty())
        die(format(_("No player number specified. Use `%s -h' for help"), progname));

    if (!gameDir)
        gameDir = ".";
    if (!rootDir)
        rootDir = gameDir;

    initDirectories(gameDir, rootDir);

    // Execute
    if (opt_dry_run) {
        LoggingSweepProcessor loggingProcessor;
        controller.processFiles(loggingProcessor);
    } else {
        controller.processFiles(processor);
        controller.processIndex();
    }

    // Status
    int numPlayers = 0;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (controller.getRemainingPlayers().contains(i))
            ++numPlayers;

    if (!opt_dry_run) {
        std::cout << format(_("%d file%!1{s%} removed.\n"), processor.getNumRemovedFiles());
        if (numPlayers == 0)
            std::cout << _("No player data remains.\n");
        else
            std::cout << format(_("%d player%1{'s%|s'%} data remains.\n"), numPlayers);
    }
    return 0;
}
