/**
  *  \file tools/c2unpack.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include "cpluslib/format.h"
#include "cpluslib/string.h"
#include "game/player.h"
#include "game/unpack.h"
#include "util/console.h"
#include "util/misc.h"
#include "util/nls.h"
#include "version.h"
#include "game/gamedirs.h"
#include "game/rst.h"
#include "util/except.h"

#define VERSION PCC2_VERSION

static const char* progname;

/** Exit with error. */
static void
die(string_t str)
{
    std::cerr << progname << ": " << str << std::endl;
    std::exit(1);
}

static void
help()
{
    std::cout << format(_("PCC2 Result File Unpacker v%s - (c) 2009 Stefan Reuther\n"), VERSION)
              << format(_("\n"
                          "Usage:\n"
                          "  %s [-h]\n"
                          "  %$0s [-watnfdx] [player number] [gamedir [rootdir]]\n\n"
                          "Options:\n"
                          "  -w       Create Windows (3.5) format [default]\n"
                          "  -d       Create DOS (3.0) format\n"
                          "  -a       Ignore version 3.5 part of RST\n"
                          "  -t       Create TARGETx.EXT files\n"
                          "  -n       Do not attempt to fix host-side errors\n"
                          "  -f       Force unpack of files with failing checksums\n"
                          "  -x       Increase verbosity\n"
                          "\n"
                          "Report bugs to <Streu@gmx.de>\n"),
                        progname);
    std::exit(0);
}

int
main(int /*argc*/, char** argv)
{
    GUnpacker theUnpacker;
    const char* gameDir = 0;
    const char* rootDir = 0;
    bool playerSetUsed = false;
    GPlayerSet players;

    progname = *argv;
    initNLS();
    enableStdioConsole();

    while (const char* arg = *++argv) {
        if (*arg == '-') {
            while (*++arg) {
                switch (*arg) {
                 case 'w': /* default */
                    theUnpacker.setFormat(theUnpacker.WindowsFormat);
                    break;
                 case 'd':
                    theUnpacker.setFormat(theUnpacker.DosFormat);
                    break;
                 case 'a':
                    theUnpacker.setIgnore35Part(true);
                    break;
                 case 't':
                    theUnpacker.setCreateTargetExt(true);
                    break;
                 case 'n':
                    theUnpacker.setFixErrors(false);
                    break;
                 case 'f':
                    theUnpacker.setForceIgnoreErrors(true);
                    break;
                 case 'x':
                    theUnpacker.setVerbose(true);
                    break;
                 case 'h':
                    help();
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        } else {
            int n;
            if (stringValue(arg, n) && n > 0 && n <= NUM_PLAYERS) {
                /* Player number */
                players += n;
                playerSetUsed = true;
            } else if (!gameDir) {
                /* Game directory */
                gameDir = arg;
            } else if (!rootDir) {
                /* Root directory */
                rootDir = arg;
            } else {
                die(_("too many arguments"));
            }
        }
    }

    if (!gameDir)
        gameDir = ".";
    if (!rootDir)
        rootDir = gameDir;

    initDirectories(gameDir, rootDir);

    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (!playerSetUsed || players.contains(i)) {
            Ptr<Stream> rst;
            string_t filename = format("player%d.rst", i);

            try {
                // Open and unpack the file
                rst = game_file_dir->openFile(filename, Stream::C_READ);
                GResultFile rstfile(*rst);
                console.write(LOG_INFO, format(_("=== Unpacking player %d... ==="), i));
                theUnpacker.unpack(rstfile, i);
            }
            catch (FileProblemException& e) {
                // If no player set was given, and the exception occurs
                // during opening the RST, it is not fatal.
                if (playerSetUsed || rst)
                    std::cerr << e.getFileName() << ": " << e.what() << "\n";
            }
            catch (std::exception& e) {
                std::cerr << filename << ": " << e.what() << "\n";
            }
        }
    }
}
